#!/usr/bin/perl

# Usage: ifgen.pl < DOM.idl

# a q&d idl parser just for parsing DOM.idl
sub parse_idl {
    my $cur_if;
    my $if_attr;

    while (<STDIN>) {
	if (!$cur_if && /^\s*interface\s+(\w+)\s*(:\s*(\w+))?\s*\{\s*$/) {
	    $cur_if = $1;
	    $if_attr = {};
	    $if_methods = [];
	    if ($3) {
		$$if_attr{'super'} = $3;
	    }
        } elsif ($cur_if && /^\s*\}\s*\;\s*$/) {
	    $$if_attr{'methods'} = $if_methods;
	    $ifs{$cur_if} = $if_attr;
	    undef $cur_if;
	} elsif ($cur_if) {
	    $line = $_;
	    $line =~ s/\/\/.*//;
	    if ($line =~ /;/) {
		$method = $prefix.' '.$line;
		push @$if_methods, $method;
		$prefix = '';
	    } else {
		chomp $line;
		$prefix .= $line;
	    }
	}
    }
}


sub dump_ifs {
foreach $if (keys %ifs) {
    print "$if: {";
	%attr = %{$ifs{$if}};
$sep = "";
foreach $key (keys %attr) {
    print "$sep$key => $attr{$key}";
    $sep = ", ";
}
print "}\n";
}
}

sub toposort {
    my (@ifs) = @_;
    my %visited;
    my @result;
    my @queue;

    @queue = sort @ifs;

    while (@queue) {
	my $node = shift @queue;
	$super = ${$ifs{$node}}{'super'};
    if ($super && !$visited{$super}) {
	unshift @queue, $node;
	unshift @queue, $super;
    } elsif (!$visited{$node}) {
	push @result, $node;
	$visited{$node} = 1;
    }
}

    return @result;
}

sub is_const {
    my ($method) = @_;

    return ($method =~ /^\s*const\b/);
}

sub is_attribute {
    my ($method) = @_;

    return ($method =~ /\battribute\b/);
}

sub is_readonly {
    my ($method) = @_;

    return ($method =~ /\breadonly\b/);
}

sub attr_name {
    my ($method) = @_;

    if ($method =~ /\s+(\w+)\s*;/) {
	return $1;
    }
}

sub attr_type {
    my ($method) = @_;

    if ($method =~ /\battribute\s+(.*\S)\s+\w+\s*;/) {
	return $1;
    }
}

sub meth_return_type {
    my ($method) = @_;
    if ($method =~ /^\s*(.*\S)\s+\w+\s*\(/) {
	return $1;
    }
}

sub meth_name {
    my ($method) = @_;
    if ($method =~ /\b(\w+)\s*\(/) {
	return $1;
    }
}

sub meth_args {
    my ($method) = @_;
    if ($method =~ /\((.*)\)/) {
	return $1;
    }
}

sub veryshort {
    my ($ifname) = @_;
    my $shortname = $ifname;

    if ($ifname eq 'Notation') {
	return 'not';
    } elsif ($ifname eq 'Element') {
	return 'el';
    } elsif ($ifname eq 'Entity') {
	return 'ent';
    } elsif ($ifname eq 'Parser') {
	return 'pars';
    } elsif ($ifname eq 'CDATASection') {
	return 'cds';
    } elsif ($ifname eq 'Document') {
	return 'doc';
    }
    $shortname =~ s/[a-z]//g;
    return lc $shortname;
}

# return value includes trailing blank if appropriate
sub dom_type_to_c {
    my ($domtype) = @_;

    if ($domtype eq 'unsigned long') {
	return 'unsigned long ';
    } elsif ($domtype eq 'void') {
	return 'void ';
    } elsif ($domtype eq 'unsigned short') {
	return 'unsigned short ';
    } elsif ($domtype eq 'boolean') {
	return 'GdomeBoolean ';
    }
    return ("Gdome$domtype *");
}

sub need_result_arg {
    my ($domtype) = @_;
    my $ctype = dom_type_to_c ($domtype);

    return 0 if ($ctype =~ / $/);
    return 0 if ($ctype eq 'GdomeDOMString *');
    return 1;
}

sub gen_typedefs {
    foreach $if (toposort (keys %ifs)) {
	print "typedef struct _Gdome$if Gdome$if;\n";
    }
    print "\n";
    foreach $if (toposort (keys %ifs)) {
	print "typedef struct _Gdome${if}Vtab Gdome${if}Vtab;\n";
    }
    print "\n";
}

sub gen_objs {
    foreach my $if (toposort (keys %ifs)) {
	print "struct _Gdome$if {\n";
	print "  Gdome${if}Vtab *vtab;\n";
#	print "  GdomeId h1, h2, h3;\n";
	print "};\n";
	print "\n";
    }
    print "\n";
}

sub gen_consts {
    my $blank;

    foreach my $if (toposort (keys %ifs)) {
	%if = %{$ifs{$if}};
	@methods = @{$if{'methods'}};
	foreach my $meth (@methods) {
	    if (is_const ($meth)) {
		# it's a method, not an attribute
		if ($meth =~ /\b(\w+)\s*\=\s*(.*);/) {
		    print "#define GDOME_$1 $2\n";
		    $blank = "\n";
		}
	    }
	}
	print $blank;
	$blank = '';
    }
}

sub gen_method {
    my ($method) = @_;
    my ($rettype, $methname, $args) = @{$method};

    print "  $rettype(*$methname) ($args);\n";
}

sub get_cmethods {
    foreach my $if (toposort (keys %ifs)) {
	%if = %{$ifs{$if}};
	@methods = @{$if{'methods'}};
	$cmeths = [];
	if (!$if{'super'}) {
	    push @{$cmeths}, ['void ', 'unref', "Gdome$if *self"];
	}
	foreach my $meth (@methods) {
	    if (is_attribute ($meth)) {
		$attrname = attr_name ($meth);
		$ctype = dom_type_to_c (attr_type ($meth));
		$args = "Gdome$if *self";
#		if (need_result_arg (attr_type ($meth))) {
#		    $args .= ", ${ctype}result";
#		}
		$cmethod = [$ctype, $attrname, $args];
		push @{$cmeths}, $cmethod;
		if (!is_readonly ($meth)) {
		    $cmethod = ['void ', "set_$attrname", "Gdome$if *self, $ctype$attrname"];
		    push @{$cmeths}, $cmethod;
		}
	    } elsif (!is_const ($meth)) {
		# it's a method, not an attribute
		$meth =~ s/\braises\s*\([^\)]*\)//;
#		print "/* $meth */\n";
		$methname = meth_name ($meth);
		$crtype = dom_type_to_c (meth_return_type ($meth));
		$args = "Gdome$if *self";
		foreach my $arg (split (/\s*,\s*/, meth_args ($meth))) {
		    $arg =~ s/in\s+//;
		    if ($arg =~ /^(.*\S)\s+(\w+)$/) {
			$argname = $2;
			$catype = dom_type_to_c ($1);
			$args .= ", $catype$argname";
		    }
		}
#		if (need_result_arg (meth_return_type ($meth))) {
#		    $args .= ", ${crtype}result";
#		}
		$cmethod = [$crtype, $methname, $args];
		push @{$cmeths}, $cmethod;
	    }
	}
	${$ifs{$if}}{'cmethods'} = $cmeths;
    }
}

sub gen_vtabs {
    foreach my $if (toposort (keys %ifs)) {
	%if = %{$ifs{$if}};
	print "struct _Gdome${if}Vtab {\n";
	if ($if{'super'}) {
	    print "  Gdome$if{'super'}Vtab super;\n";
	}
	@cmethods = @{$if{'cmethods'}};
	foreach my $cmethod (@cmethods) {
	    gen_method ($cmethod);
	}
#	print "\n";
#	print "  void *private;\n";
	print "};\n";
	print "\n";
    }
}

sub gen_invoker_headers {
    my $blank;
    foreach my $if (toposort (keys %ifs)) {
	%if = %{$ifs{$if}};
	@cmethods = @{$if{'cmethods'}};
	foreach my $cmethod (@cmethods) {
	    my ($rettype, $methname, $args) = @{$cmethod};
	    my $short = veryshort ($if);
	    print "${rettype}gdome_${short}_$methname ($args);\n";
	    $blank = "\n";
	}
	print $blank;
	$blank = '';
    }
}

sub gen_invokers {
    foreach my $if (toposort (keys %ifs)) {
	%if = %{$ifs{$if}};
	@cmethods = @{$if{'cmethods'}};
	foreach my $cmethod (@cmethods) {
	    my ($rettype, $methname, $args) = @{$cmethod};
	    $rettype =~ s/ $//;
	    print "$rettype\n";
	    my $short = veryshort ($if);
	    print "gdome_${short}_$methname ($args)\n";
	    print "{\n";
	    if ($rettype eq 'void') {
		$return = '';
	    } else {
		$return = 'return ';
	    }
	    print "  ${return}self->vtab->$methname (";
	    $sep = '';
	    foreach my $arg (split (/\s*,\s*/, $args)) {
		if ($arg =~ /\W(\w+)$/) {
		    print "$sep$1";
		    $sep = ', ';
		}
	    }
	    print ");\n";
	    print "}\n";
	    print "\n";
	}
    }
}

%libxml_name = (
    Node => 'xmlNode',
    Notation => 'xmlNotation',
    Attr => 'xmlAttr',
    Document => 'xmlDoc',
    DocumentType => 'xmlDtd'
		);


sub gen_priv_typedefs {
    my ($impl) = @_;

    foreach $if (toposort (keys %ifs)) {
	print "typedef struct _Gdome_${impl}_$if Gdome_${impl}_$if;\n";
    }
    print "\n";
}

# returns ('xmlNode', 'n');
sub find_privates {
    my ($if) = @_;

    for (my $klass = $if; $klass; $klass = $super) {
	if ($libxml_name{$klass}) {
	    $ksh = veryshort ($klass);
	    return ($libxml_name{$klass}, $ksh);
	} else {
	    %kif = %{$ifs{$klass}};
	    $super = $kif{'super'};
	}
    }
    return ();
}

sub gen_priv_objs {
    my ($impl) = @_;
    my ($itype, $iname);

    foreach $if (toposort (keys %ifs)) {
	print "struct _Gdome_${impl}_$if {\n";
	my $short = veryshort ($if);
	print "  Gdome$if super;\n";
	($itype, $iname) = find_privates ($if);
	if ($itype) {
	    print "  $itype *$iname;\n";
	}
	print "};\n";
	print "\n";
    }
}

sub gen_skeletons {
    my ($impl) = @_;

    foreach my $if (toposort (keys %ifs)) {
	print "/* implementations of methods in interface $if */\n";
	print "\n";
	%if = %{$ifs{$if}};
	@cmethods = @{$if{'cmethods'}};
	if ($libxml_name{$if} && $if{'super'}) {
	    %sif = %{$ifs{$if{'super'}}};
	    unshift @cmethods, @{$sif{'cmethods'}};
	}
	foreach my $cmethod (@cmethods) {
	    my ($rettype, $methname, $args) = @{$cmethod};
	    $rettype =~ s/ $//;
	    print "static $rettype\n";
	    $short = veryshort ($if);
	    print "gdome_${impl}_${short}_$methname ($args)\n";
	    print "{\n";
	    print "  Gdome_${impl}_$if *priv = (Gdome_${impl}_$if *)self;\n";
	    my ($itype, $iname) = find_privates ($if);
	    if ($itype) {
		print "  $itype *$iname = priv->$iname;\n";
	    }
	    %kif = %{$ifs{$klass}};
	    print "\n";
#	    if ($libxml_name{$if}) {
#		$lxn = $libxml_name{$if};
#		print "  $lxn *$short = ($lxn *)self->h1.p;\n";
#		print "\n";
#	    }
	    if ($rettype =~ /\*/) {
		print "  return NULL;\n";
	    } elsif ($rettype ne 'void') {
		print "  return 0;\n";
	    }
	    print "}\n";
	    print "\n";
	}
    }
}

sub gen_generic_vtab_contents {
    my ($if, $short, $indent) = @_;
    my %if = %{$ifs{$if}};
    my $nl = '';
    my $child_short;

    if ($if{'super'}) {
	print (("  " x $indent)."{\n");
	if ($libxml_name{$if}) {
	    $child_short = $short;
	} else {
	    $child_short = veryshort ($if{'super'});
	}
	$nl = gen_generic_vtab_contents ($if{'super'}, $child_short, $indent + 1);
	print "\n";
	print (("  " x $indent)."}");
	$nl = ",\n";
    }
    my @cmethods = @{$if{'cmethods'}};
    foreach my $cmethod (@cmethods) {
	my ($rettype, $methname, $args) = @{$cmethod};
	print ($nl.("  " x $indent)."gdome_xml_${short}_$methname");
	$nl = ",\n";
    }
    return $nl;
}

sub gen_generic_vtabs {
    my ($impl) = @_;
    my $nl = '';
    foreach my $if (toposort (keys %ifs)) {
	my %if = %{$ifs{$if}};
	$short = veryshort ($if);
	print "const Gdome${if}Vtab gdome_${impl}_${short}_vtab = {\n";
	$nl = gen_generic_vtab_contents ($if, $short, 1);
	print "\n";
#	print "\n";
#	print "  NULL\n";
	print "};\n";
	print "\n";
    }
}

sub gen_generic_vtab_decls {
    my ($impl) = @_;
    my $blank;
#    print "/* Vtab decls */\n";
#    print "\n";
    foreach my $if (toposort (keys %ifs)) {
	my %if = %{$ifs{$if}};
	$short = veryshort ($if);
	print "const Gdome${if}Vtab gdome_${impl}_${short}_vtab;\n";
	$blank = "\n";
    }
    print $blank;
    $blank = '';
}

parse_idl ();
#dump_ifs ();
print "----- gdome.h -----\n";
print << 'EOF';
typedef struct _GdomeDOMString GdomeDOMString;
typedef int GdomeBoolean;

struct _GdomeDOMString {
  void (*unref) (GdomeDOMString *self);
  char *str;
};

EOF
gen_typedefs ();
gen_consts ();
gen_objs ();
get_cmethods ();
gen_vtabs ();
print << 'EOF';
void gdome_str_unref (GdomeDOMString *self);

EOF
gen_invoker_headers ();
print "----- gdome.c -----\n";
print << 'EOF';
#include "gdome.h"

void gdome_str_unref (GdomeDOMString *self)
{
  self->unref (self);
}

EOF
gen_invokers ();
print "----- gdome-xml.c -----\n";
print << 'EOF';
#include "gdome.h"
#include "gnome-xml/tree.h"
#include "gdome-xml.h"

EOF
$impl = 'xml';
gen_priv_typedefs ($impl);
gen_priv_objs ($impl);
gen_generic_vtab_decls ($impl);
gen_skeletons ($impl);
gen_generic_vtabs ($impl);
